@@ -334,6 +334,7 @@ bool parse_message
 		std::cerr << "ERROR: wrong type of ASCII Armor found (type = " << (int)atype << ")" << std::endl;
 		return false;
 	}
+	size_t pnum = 0;
 	while (pkts.size() && ptag)
 	{
 		tmcg_octets_t pkesk_keyid;
@@ -343,17 +344,23 @@ bool parse_message
 		std::vector<std::string> capl;
 		std::vector< std::vector<gcry_mpi_t> > c_ik;
 		ptag = CallasDonnerhackeFinneyShawThayerRFC4880::PacketDecode(pkts, ctx, current_packet, qual, capl, v_i, c_ik);
+		++pnum;
 		if (opt_verbose)
-			std::cout << "PacketDecode() = " << (int)ptag;
-		if (!ptag)
+			std::cout << "PacketDecode() = " << (int)ptag << " version = " << (int)ctx.version << std::endl;
+		if (ptag == 0x00)
 		{
-			std::cerr << "ERROR: parsing OpenPGP packets failed at position " << pkts.size() << std::endl;
+			std::cerr << "ERROR: parsing OpenPGP packets failed at #" << pnum << " and position " << pkts.size() << std::endl;
 			cleanup_ctx(ctx);
 			cleanup_containers(qual, v_i, c_ik);
-			return false;
+			return false; // parsing error detected
+		}
+		else if (ptag == 0xFE)
+		{
+			std::cerr << "WARNING: unrecognized OpenPGP packet found at #" << pnum << " and position " << pkts.size() << std::endl;
+			cleanup_ctx(ctx);
+			cleanup_containers(qual, v_i, c_ik);
+			continue; // ignore packet
 		}
-		if (opt_verbose)
-			std::cout << " tag = " << (int)ptag << " version = " << (int)ctx.version << std::endl;
 		switch (ptag)
 		{
 			case 1: // Public-Key Encrypted Session Key
@@ -361,7 +368,7 @@ bool parse_message
 					std::cout << " pkalgo = " << (int)ctx.pkalgo << std::endl;
 				if (ctx.pkalgo != 16)
 				{
-					std::cerr << "WARNING: public-key algorithm not supported; packet ignored" << std::endl;
+					std::cerr << "WARNING: public-key algorithm not supported; packet #" << pnum << " ignored" << std::endl;
 					break;
 				}
 				if (opt_verbose)
@@ -422,7 +429,7 @@ bool parse_message
 				}
 				break;
 			default:
-				std::cerr << "ERROR: unrecognized OpenPGP packet found" << std::endl;
+				std::cerr << "ERROR: unexpected OpenPGP packet " << (int)ptag<< " found at #" << pnum << std::endl;
 				cleanup_ctx(ctx);
 				cleanup_containers(qual, v_i, c_ik);
 				return false;
@@ -491,17 +498,17 @@ bool decrypt_message
 		return false;
 	}
 	gcry_error_t ret;
-	tmcg_octets_t prefix, litmdc;
+	tmcg_octets_t prefix, pkts;
 	if (have_seipd)
-		ret = CallasDonnerhackeFinneyShawThayerRFC4880::SymmetricDecrypt(in, key, prefix, false, symalgo, litmdc);
+		ret = CallasDonnerhackeFinneyShawThayerRFC4880::SymmetricDecrypt(in, key, prefix, false, symalgo, pkts);
 	else
-		ret = CallasDonnerhackeFinneyShawThayerRFC4880::SymmetricDecrypt(in, key, prefix, true, symalgo, litmdc);
+		ret = CallasDonnerhackeFinneyShawThayerRFC4880::SymmetricDecrypt(in, key, prefix, true, symalgo, pkts);
 	if (ret)
 	{
 		std::cerr << "ERROR: SymmetricDecrypt() failed" << std::endl;
 		return false;
 	}
-	// parse content of decrypted message
+	// parse the content of decrypted message
 	tmcg_openpgp_packet_ctx ctx;
 	std::vector<gcry_mpi_t> qual, v_i;
 	std::vector<std::string> capl;
@@ -509,27 +516,40 @@ bool decrypt_message
 	bool have_lit = false, have_mdc = false;
 	tmcg_octets_t lit, mdc_hash;
 	tmcg_byte_t ptag = 0xFF;
-	if (litmdc.size() > (sizeof(ctx.mdc_hash) + 2))
-		lit.insert(lit.end(), litmdc.begin(), litmdc.end() - (sizeof(ctx.mdc_hash) + 2));
-	while (litmdc.size() && ptag)
+	size_t pnum = 0, mdc_len = sizeof(ctx.mdc_hash) + 2;
+	if (pkts.size() > mdc_len)
+		lit.insert(lit.end(), pkts.begin(), pkts.end() - mdc_len); // store literal data
+	while (pkts.size() && ptag)
 	{
 		tmcg_octets_t current_packet;
-		ptag = CallasDonnerhackeFinneyShawThayerRFC4880::PacketDecode(litmdc, ctx, current_packet, qual, capl, v_i, c_ik);
+		ptag = CallasDonnerhackeFinneyShawThayerRFC4880::PacketDecode(pkts, ctx, current_packet, qual, capl, v_i, c_ik);
+		++pnum;
 		if (opt_verbose)
-			std::cout << "PacketDecode() = " << (int)ptag;
-		if (!ptag)
+			std::cout << "PacketDecode() = " << (int)ptag << " version = " << (int)ctx.version << std::endl;
+		if (ptag == 0x00)
 		{
-			std::cerr << "ERROR: parsing OpenPGP packets failed at position " << litmdc.size() << std::endl;
+			std::cerr << "ERROR: parsing OpenPGP packets failed at #" << pnum << " and position " << pkts.size() << std::endl;
 			cleanup_ctx(ctx);
 			cleanup_containers(qual, v_i, c_ik);
-			return false;
+			return false; // parsing error detected
+		}
+		else if (ptag == 0xFE)
+		{
+			std::cerr << "WARNING: unrecognized OpenPGP packet found at #" << pnum << " and position " << pkts.size() << std::endl;
+			cleanup_ctx(ctx);
+			cleanup_containers(qual, v_i, c_ik);
+			continue; // ignore packet
 		}
-		if (opt_verbose)
-			std::cout << " tag = " << (int)ptag << " version = " << (int)ctx.version << std::endl;
 		switch (ptag)
 		{
+			case 2: // Signature
+				std::cerr << "WARNING: signature OpenPGP packet found; not supported and ignored" << std::endl;
+				break;
+			case 4: // One-Pass Signature
+				std::cerr << "WARNING: one-pass signature OpenPGP packet found; not supported and ignored" << std::endl;
+				break;
 			case 8: // Compressed Data
-				std::cerr << "WARNING: compressed OpenPGP packet found; not supported" << std::endl;
+				std::cerr << "WARNING: compressed OpenPGP packet found; not supported and ignored" << std::endl;
 				break;
 			case 11: // Literal Data
 				if (!have_lit)
@@ -554,7 +574,7 @@ bool decrypt_message
 					mdc_hash.push_back(ctx.mdc_hash[i]);
 				break;
 			default:
-				std::cerr << "ERROR: unrecognized OpenPGP packet found" << std::endl;
+				std::cerr << "ERROR: unexpected OpenPGP packet " << (int)ptag<< " found at #" << pnum << std::endl;
 				cleanup_ctx(ctx);
 				cleanup_containers(qual, v_i, c_ik);
 				return false;
@@ -606,6 +626,7 @@ bool parse_signature
 		std::cerr << "ERROR: wrong type of ASCII Armor found (type = " << (int)atype << ")" << std::endl;
 		return false;
 	}
+	size_t pnum = 0;
 	while (pkts.size() && ptag)
 	{
 		tmcg_openpgp_packet_ctx ctx;
@@ -614,13 +635,23 @@ bool parse_signature
 		std::vector<std::string> capl;
 		std::vector< std::vector<gcry_mpi_t> > c_ik;
 		ptag = CallasDonnerhackeFinneyShawThayerRFC4880::PacketDecode(pkts, ctx, current_packet, qual, capl, v_i, c_ik);
-		if (!ptag)
+		++pnum;
+		if (opt_verbose)
+			std::cout << "PacketDecode() = " << (int)ptag << " version = " << (int)ctx.version << std::endl;
+		if (ptag == 0x00)
 		{
-			std::cerr << "ERROR: parsing OpenPGP packets failed at position " << pkts.size() << std::endl;
+			std::cerr << "ERROR: parsing OpenPGP packets failed at #" << pnum << " and position " << pkts.size() << std::endl;
 			cleanup_ctx(ctx);
 			cleanup_containers(qual, v_i, c_ik);
 			return false; // parsing error detected
 		}
+		else if (ptag == 0xFE)
+		{
+			std::cerr << "WARNING: unrecognized OpenPGP packet found at #" << pnum << " and position " << pkts.size() << std::endl;
+			cleanup_ctx(ctx);
+			cleanup_containers(qual, v_i, c_ik);
+			continue; // ignore packet
+		}
 		switch (ptag)
 		{
 			case 2: // Signature Packet
@@ -679,6 +710,11 @@ bool parse_signature
 					std::cerr << "WARNING: more than one admissible signature; packet ignored" << std::endl;
 				}
 				break;
+			default:
+				std::cerr << "ERROR: unexpected OpenPGP packet " << (int)ptag<< " found at #" << pnum << std::endl;
+				cleanup_ctx(ctx);
+				cleanup_containers(qual, v_i, c_ik);
+				return false;
 		}
 		// cleanup allocated buffers and mpi's
 		cleanup_ctx(ctx);
@@ -727,7 +763,9 @@ bool parse_public_key
 		std::vector< std::vector<gcry_mpi_t> > c_ik;
 		ptag = CallasDonnerhackeFinneyShawThayerRFC4880::PacketDecode(pkts, ctx, current_packet, qual, capl, v_i, c_ik);
 		++pnum;
-		if (!ptag)
+		if (opt_verbose)
+			std::cout << "PacketDecode() = " << (int)ptag << " version = " << (int)ctx.version << std::endl;
+		if (ptag == 0x00)
 		{
 			std::cerr << "ERROR: parsing OpenPGP packets failed at #" << pnum << " and position " << pkts.size() << std::endl;
 			gcry_mpi_release(dsa_r);
@@ -736,7 +774,14 @@ bool parse_public_key
 			gcry_mpi_release(elg_s);
 			cleanup_ctx(ctx);
 			cleanup_containers(qual, v_i, c_ik);
-			return false;
+			return false; // parsing error detected
+		}
+		else if (ptag == 0xFE)
+		{
+			std::cerr << "WARNING: unrecognized OpenPGP packet found at #" << pnum << " and position " << pkts.size() << std::endl;
+			cleanup_ctx(ctx);
+			cleanup_containers(qual, v_i, c_ik);
+			continue; // ignore packet
 		}
 		switch (ptag)
 		{
@@ -747,7 +792,7 @@ bool parse_public_key
 				if (pubdsa && !subelg && !uid && !uat && (ctx.type >= 0x10) && (ctx.type <= 0x13) && 
 					CallasDonnerhackeFinneyShawThayerRFC4880::OctetsCompare(keyid, issuer))
 				{
-					std::cerr << "ERROR: no uid found for this self-signature" << std::endl;
+					std::cerr << "ERROR: no uid/uat found for this self-signature" << std::endl;
 					gcry_mpi_release(dsa_r);
 					gcry_mpi_release(dsa_s);
 					gcry_mpi_release(elg_r);
@@ -782,6 +827,8 @@ bool parse_public_key
 					dsa_hspd.clear();
 					for (size_t i = 0; i < ctx.hspdlen; i++)
 						dsa_hspd.push_back(ctx.hspd[i]);
+					if (opt_verbose)
+						std::cout << "INFO: dsa_hspd.size() = " << dsa_hspd.size() << std::endl;
 					if (dsa_pkalgo != 17)
 					{
 						std::cerr << "ERROR: public-key signature algorithms other than DSA not supported" << std::endl;
@@ -795,6 +842,11 @@ bool parse_public_key
 					}
 					gcry_mpi_set(dsa_r, ctx.r);
 					gcry_mpi_set(dsa_s, ctx.s);
+					unsigned int rbits = 0, sbits = 0;
+					rbits = gcry_mpi_get_nbits(dsa_r);
+					sbits = gcry_mpi_get_nbits(dsa_s);
+					if (opt_verbose)
+						std::cout << "INFO: rbits = " << rbits << " sbits = " << sbits << std::endl;
 					if ((dsa_hashalgo < 8) || (dsa_hashalgo >= 11))
 						std::cerr << "WARNING: insecure hash algorithm " << (int)dsa_hashalgo << " used for signatures" << std::endl;
 					time_t kmax = dsa_creation + ctx.keyexpirationtime;
@@ -806,7 +858,7 @@ bool parse_public_key
 					for (size_t i = 0; i < current_packet.size(); i++)
 						uidsig.push_back(current_packet[i]);
 				}
-				else if (pubdsa && !subelg && uid && uat && (ctx.type >= 0x10) && (ctx.type <= 0x13) && 
+				else if (pubdsa && !subelg && !uid && uat && (ctx.type >= 0x10) && (ctx.type <= 0x13) && 
 					CallasDonnerhackeFinneyShawThayerRFC4880::OctetsCompare(keyid, issuer))
 				{
 					std::cerr << "WARNING: ignore certifying self-signature for a user attribute" << std::endl;
@@ -835,6 +887,8 @@ bool parse_public_key
 					elg_hspd.clear();
 					for (size_t i = 0; i < ctx.hspdlen; i++)
 						elg_hspd.push_back(ctx.hspd[i]);
+					if (opt_verbose)
+						std::cout << "INFO: elg_hspd.size() = " << elg_hspd.size() << std::endl;
 					if (elg_pkalgo != 17)
 					{
 						std::cerr << "ERROR: public-key signature algorithms other than DSA not supported" << std::endl;
@@ -848,6 +902,11 @@ bool parse_public_key
 					}
 					gcry_mpi_set(elg_r, ctx.r);
 					gcry_mpi_set(elg_s, ctx.s);
+					unsigned int rbits = 0, sbits = 0;
+					rbits = gcry_mpi_get_nbits(elg_r);
+					sbits = gcry_mpi_get_nbits(elg_s);
+					if (opt_verbose)
+						std::cout << "INFO: rbits = " << rbits << " sbits = " << sbits << std::endl;
 					if ((elg_hashalgo < 8) || (elg_hashalgo >= 11))
 						std::cerr << "WARNING: insecure hash algorithm " << (int)elg_hashalgo << " used for signatures" << std::endl;
 					time_t kmax = elg_creation + ctx.keyexpirationtime;
@@ -918,7 +977,7 @@ bool parse_public_key
 			case 13: // User ID Packet
 				if (uid)
 					std::cerr << "WARNING: more than one uid packet found; using last to verify signature" << std::endl;
-				uid = true;
+				uid = true, uat = false;
 				userid = "";
 				for (size_t i = 0; i < sizeof(ctx.uid); i++)
 				{
@@ -933,12 +992,12 @@ bool parse_public_key
 				{
 					if (subelg)
 						std::cerr << "WARNING: ElGamal subkey already found; the last one is used" << std::endl;
-					subelg = true, sigelg = false;
+					subelg = true, sigelg = false, sigelgV3 = false;
 					gcry_mpi_set(elg_p, ctx.p);
 					gcry_mpi_set(elg_g, ctx.g);
 					gcry_mpi_set(elg_y, ctx.y);
 					elg_creation = ctx.keycreationtime;
-					keycreationtime_out = ctx.keycreationtime; // FIXME: overwrites creation time of primary key
+					keycreationtime_out = ctx.keycreationtime; // FIXME: this overwrites saved creation time of primary key
 					sub.clear();
 					CallasDonnerhackeFinneyShawThayerRFC4880::PacketSubEncode(ctx.keycreationtime, ctx.pkalgo,
 						elg_p, dsa_q, elg_g, elg_y, sub);
@@ -960,11 +1019,13 @@ bool parse_public_key
 				break;
 			case 17: // User Attribute Packet
 				std::cerr << "WARNING: user attribute packet found; ignored" << std::endl;
-				uat = true;
-				break;
-			case 0xFE: // unrecognized packet content
-				std::cerr << "WARNING: unrecognized content of packet #" << pnum << " at position " << pkts.size() << std::endl;
+				uid = false, uat = true;
 				break;
+			default:
+				std::cerr << "ERROR: unexpected OpenPGP packet found at #" << pnum << " and position " << pkts.size() << std::endl;
+				cleanup_ctx(ctx);
+				cleanup_containers(qual, v_i, c_ik);
+				return false;
 		}
 		// cleanup allocated buffers and mpi's
 		cleanup_ctx(ctx);
@@ -1044,7 +1105,7 @@ bool parse_public_key
 		if ((flags & 0x80) == 0x80)
 			std::cout << "M"; // The private component of this key may be in the possession of more than one person.
 		std::cout << std::endl;
-		std::cout << "INFO: userid = " << userid << std::endl;
+		std::cout << "INFO: userid = \"" << userid << "\"" << std::endl;
 		std::cout << "INFO: dsa_sigtype = 0x" << std::hex << (int)dsa_sigtype << std::dec << 
 			" dsa_pkalgo = " << (int)dsa_pkalgo << " dsa_hashalgo = " << (int)dsa_hashalgo << " dsa_hspd.size() = " << dsa_hspd.size() << std::endl;
 	}
